from jwt import InvalidTokenError



class AAsuError(Exception):
    """Base class for all AAsu exceptions."""
    pass



class JwtDenied(InvalidTokenError):
    """Raised only when either verify_token or verify_data fails."""
    pass



from typing import Any, Self, Type, TypeVar, Generic, overload, cast

from pydantic import BaseModel
from pymongo import MongoClient
from pymongo.cursor import Cursor as PyMongoCursor
from pymongo.command_cursor import CommandCursor as PyMongoCommandCursor


__all__ = [
    "Database", "Collection", "Cursor", "AggregateCursor",
    "ASC", "DESC"
]


ColModelT = TypeVar("ColModelT")
ASC = 1
DESC = -1



class Database:

    def __init__(self,
                 uri: str,
                 db_name: str,
                 *,
                 mongo_client: MongoClient | None = None):
        """Initialize a database connection with a given
        MongoDB URL or client"""
        if mongo_client is not None:
            self.client = mongo_client
        else:
            self.client = MongoClient(uri)

        self.db = self.client[db_name]
        """The pymongo database instance (used in internal)"""
        self._collections: dict[str, Collection[Any]] = {}


    @overload
    def collection(
        self,
        name: str,
        model: Type[ColModelT],
        *,
        primary_key: str = "id"
    ) -> "Collection[ColModelT]":
        ...

    @overload
    def collection(
        self,
        name: str,
        model: None = None,
        *,
        primary_key: str = "id"
    ) -> "Collection[dict[str, Any]]":
        ...


    def collection(self,
                   name: str,
                   model: Type[ColModelT] | None = None,
                   *,
                   primary_key: str = "id") -> "Collection[ColModelT]":
        """Creates a collection/table of data  following
        a given Pydantic Model"""
        if name in self._collections:
            return self._collections[name]

        col = Collection(name, self, model=model, primary_key=primary_key)
        self._collections[name] = col
        return col
    

    def __getitem__(self, name: str) -> "Collection":
        return self.collection(name)



class Collection(Generic[ColModelT]):

    def __init__(self, 
                 name: str,
                 database: Database,
                 model: Type[ColModelT] | None = None,
                 *,
                 primary_key: str = "id"):
        """Initialize a collection of data from a database.
        Prefer using db.collection() function instead."""
        self.name = name
        self.database = database
        self.model = model
        self.primary_key = primary_key

    @property
    def collection(self):
        """The pymongo collection instance (used in internal)"""
        return self.database.db[self.name]


    def insert(self, *objs: ColModelT | dict[str, Any]) -> None:
        """Insert one or many objects into the collection"""
        dict_objs = []
        for obj in objs:
            if isinstance(obj, BaseModel):
                dict_objs.append(obj.model_dump(mode="json"))
            else:
                dict_objs.append(obj)

        self.collection.insert_many(dict_objs)


    def find(self,
             filters: dict[str, Any] = {},
             *,
             limit: int | None = None,
             sort: list[tuple[str, int]] | None = None,
             skip: int | None = None,
             projection: dict[str, Any] | None = None) -> "Cursor[ColModelT]":
        """Finds objects in the collection.
        Actually creates a cursor"""
        cursor = Cursor(
            self, filters.copy(),
            limit=limit, projection=projection,
            skip_offset=skip or 0,
            sort_data=sort
        )
        return cursor


    def insert_or_update(self, obj: ColModelT | dict[str, Any]) -> None:
        """Insert or update a object into the collection.
        The object is updated if the primary key value already exists."""
        if isinstance(obj, BaseModel):
            dict_obj = obj.model_dump(mode="json")
        else:
            dict_obj = cast(dict[str, Any], obj)

        primarykey = dict_obj[self.primary_key]
        filters = {self.primary_key: primarykey}

        exists = self.count(filters) > 0

        if exists:
            self.update_one(filters, {"$set": dict_obj})
        else:
            self.insert(dict_obj)


    def update(self,
               filters: dict[str, Any] | str,
               update_data: dict[str, Any]) -> None:
        """Update objects in the collection that match the filters
        The filters can be a dict of filters or a primary key value"""
        ftrs = filters if isinstance(filters, dict) else {self.primary_key: filters}
        self.collection.update_many(ftrs, update_data)


    def update_one(self,
                   filters: dict[str, Any] | str,
                   update_data: dict[str, Any],
                   *,
                   sort: list[tuple[str, int]] | dict[str, int] | None = None) -> None:
        """Update a single object in the collection that matches the filters
        The filters can be a dict of filters or a primary key value"""
        ftrs = filters if isinstance(filters, dict) else {self.primary_key: filters}
        sortt = dict(sort) if isinstance(sort, list) else sort
        self.collection.update_one(ftrs, update_data, sort=sortt)


    def aggregate(self, pipeline: list[dict[str, Any]]) -> "AggregateCursor":
        """Perform aggregation pieplines inside the collection"""
        cursor = AggregateCursor(self, pipeline)
        return cursor


    def find_one(self,
                 filters: dict[str, Any],
                 projection: dict[str, Any] | None = None) -> "ColModelT | None":
        """Returns the first object that matches the filters"""
        cursor = self.find(filters, limit=1, projection=projection)
        return next(cursor, None)


    def get(self, id: Any) -> "ColModelT | None":
        """Returns the object with the given primary key value"""
        return self.find_one({self.primary_key: id})


    def count(self, filters: dict[str, Any] = {}) -> int:
        """Returns the number of objects that match the filters"""
        return self.collection.count_documents(filters)


    def delete(self, filters: dict[str, Any] | str) -> None:
        """Delete objects in the collection that match the filters
        The filters can be a dict of filters or a primary key value"""
        ftrs = filters if isinstance(filters, dict) else {self.primary_key: filters}
        self.collection.delete_many(ftrs)


    def __delitem__(self, id: Any) -> None:
        self.delete(id)



class Cursor(Generic[ColModelT]):

    def __init__(self,
                 collection: Collection[ColModelT],
                 filters: dict[str, Any],
                 limit: int | None = None,
                 projection: dict[str, Any] | None = None,
                 skip_offset: int = 0,
                 sort_data: list[tuple[str, int]] | None = None):
        self.collection = collection
        self.filters = filters
        self.record_limit = limit
        self.projection = projection
        self.skip_offset = skip_offset
        self.current_cursor: PyMongoCursor | None = None
        self.sort_data = sort_data or []


    def copy(self) -> "Cursor":
        return Cursor(
            collection=self.collection,
            filters=self.filters.copy(),
            limit=self.record_limit,
            projection=(self.projection or {}).copy(),
            skip_offset=self.skip_offset
        )


    def filter(self, filters: dict[str, Any]) -> Self:
        self.filters.update(filters)
        return self


    def limit(self, limit: int) -> Self:
        self.record_limit = limit
        return self
    

    def skip(self, skip: int) -> Self:
        self.skip_offset = skip
        return self


    def project(self, projection: dict[str, Any]) -> Self:
        self.projection = projection
        return self


    def sort(self, *sort: tuple[str, int], **sortkw: int) -> Self:
        self.sort_data = list(sort) + list(sortkw.items())
        return self


    def _build_cursor(self) -> PyMongoCursor:
        cursor = self.collection.collection.find(
            self.filters, projection=self.projection
        )

        if self.record_limit is not None:
            cursor = cursor.limit(self.record_limit)

        if self.skip_offset > 0:
            cursor = cursor.skip(self.skip_offset)

        if self.sort_data:
            cursor = cursor.sort(self.sort_data)

        self.current_cursor = cursor
        return cursor


    def first(self) -> ColModelT | None:
        obj = next(iter(self), None)
        return obj


    def all(self) -> list[ColModelT]:
        return list(iter(self))


    def __iter__(self):
        cc = self._build_cursor()
        for obj in cc:
            yield self._deserialize(obj)


    def __next__(self) -> "ColModelT":
        if self.current_cursor is None:
            self._build_cursor()

        assert self.current_cursor is not None
        obj = next(self.current_cursor)
        return self._deserialize(obj)


    def _deserialize(self, obj: dict[str, Any]) -> ColModelT:
        if self.collection.model is not None:
            assert issubclass(self.collection.model, BaseModel)
            return self.collection.model.model_validate(obj)
        else:
            return obj # type: ignore



class AggregateCursor:

    def __init__(self,
                 collection: Collection[ColModelT],
                 pipeline: list[dict[str, Any]]) -> None:
        self.collection = collection
        self.pipeline = pipeline
        self.current_cursor: PyMongoCommandCursor | None = None


    def __iter__(self):
        if self.current_cursor is None:
            self._build_cursor()

        return self


    def __next__(self):
        if self.current_cursor is None:
            self._build_cursor()

        assert self.current_cursor is not None
        obj = next(self.current_cursor)
        return obj


    def add_line(self, *pipeline: dict[str, Any]) -> Self:
        self.pipeline.extend(pipeline)
        return self


    def _build_cursor(self) -> PyMongoCommandCursor:
        cursor = self.collection.collection.aggregate(self.pipeline)
        self.current_cursor = cursor
        return cursor



from pydantic import BaseModel
from fastapi.encoders import jsonable_encoder


__all__ = ["APIModel", "apiserialize"]



class APIModel(BaseModel):

    def apiserialize(self, privacy: str | None = None):
        return self.model_dump(mode="json")



def apiserialize(obj: APIModel | BaseModel | dict | list,
                 privacy: str | None = None):
    """Serializes the given object to a JSON-serializable format, using the
    apiserialize or model_dump method"""
    if isinstance(obj, APIModel):
        l1 = obj.apiserialize(privacy=privacy)

    elif isinstance(obj, BaseModel):
        l1 = obj.model_dump(mode="json")

    else:
        l1 = obj

    secure_layer = jsonable_encoder(l1, custom_encoder={
        APIModel: lambda o: apiserialize(o, privacy=privacy)
    })
    return secure_layer



from .auth import JwtAuthConfig, JwtAuthenticator
from .cache import CacheDatabase, CacheController
from .database import Database, Collection, ASC, DESC
from .objects import APIModel, apiserialize
from .fastapicompatible import FastAPICompatibleJSONResponse, useapp



from typing import Any
from json import dumps

from fastapi import FastAPI
from fastapi.responses import JSONResponse
from fastapi.datastructures import DefaultPlaceholder

from .objects import APIModel, apiserialize


__all__ = ["FastAPICompatibleJSONResponse", "useapp"]



class FastAPICompatibleJSONResponse(JSONResponse):

    def render(self, content: Any) -> bytes:
        return super().render(apiserialize(content))


def compatible_renderer(self, content: Any) -> bytes:
    content = apiserialize(content)
    return super(self).render(content)


def patch_encoder():
    import fastapi.encoders as _encoders
    import fastapi.routing as _routing
    import fastapi.dependencies.utils as _dep_utils

    old = _encoders.jsonable_encoder

    def new(obj, **kwargs):
        if isinstance(obj, APIModel):
            return apiserialize(obj)
        return old(obj, **kwargs)

    # Also patch all modules that captured a direct reference at import time
    _encoders.jsonable_encoder = new
    _routing.jsonable_encoder = new # type: ignore

    try:
        _dep_utils.jsonable_encoder = new
    except AttributeError:
        pass


def useapp(app: FastAPI) -> FastAPI:
    respclass = app.router.default_response_class
    if isinstance(respclass, DefaultPlaceholder):
        app.router.default_response_class = FastAPICompatibleJSONResponse

    elif issubclass(respclass, JSONResponse):
        app.router.default_response_class.render = compatible_renderer

    patch_encoder()
    return app



from typing import TypeVar, Generic, Any, overload, cast
from datetime import datetime

from pydantic import BaseModel
from fastapi import FastAPI
from redis import Redis


__all__ = ["CacheDatabase", "CacheController"]


CacheControllerModelT = TypeVar("CacheControllerModelT")



class CacheDatabase:

    def __init__(self, rclient: Redis) -> None:
        self.rclient = rclient
        """The Redis client"""
        self.app: FastAPI | None = None
        """The FastAPI application (not used)"""
        self._controllers: list[CacheController] = []


    @overload
    def cacher(self,
               key: str,
               model: type[CacheControllerModelT],
               *,
               default_expiration: int | None = None) -> "CacheController[CacheControllerModelT]":
        ...

    @overload
    def cacher(self,
               key: str,
               model: None = None,
               *,
               default_expiration: int | None = None) -> "CacheController[str]":
        ...

    def cacher(self,
               key: str,
               model = None,
               *,
               default_expiration: int | None = None):
        controller = CacheController(self, key, model=model, default_expiration=default_expiration)
        self._controllers.append(controller)
        return controller




class CacheController(Generic[CacheControllerModelT]):

    def __init__(self,
                 cdb: CacheDatabase,
                 key: str,
                 *,
                 model: CacheControllerModelT | None = None,
                 default_expiration: int | None = None) -> None:
        self.cdb = cdb
        """The database"""
        self.key = key
        """The prefix used to store objects"""
        self.model = model
        """The Pydantic model of the objects you want to cache"""
        self.default_expiration = default_expiration
        """A default expiration (in seconds) for every stored object"""


    def __getitem__(self, key: str):
        return self.get(key)


    def __setitem__(self, key: str, value):
        self.set(key, value)


    def __delitem__(self, key: str) -> None:
        self.delete(key)


    def _deserialize(self, rawval: str | None) -> Any:
        if rawval is None:
            return None

        if self.model:
            obj = self.model.model_validate_json(rawval) # type: ignore
        else:
            obj = rawval

        return obj # type: ignore


    def _serialize_value(self, value: Any) -> str:
        if isinstance(value, BaseModel):
            valdata = value.model_dump_json()

        elif isinstance(value, str):
            valdata = value

        else:
            raise TypeError("value should either be a string, or a pydantic model instance")
        
        return valdata


    def set(self,
            key: str,
            value: CacheControllerModelT,
            expires_in: int | None | str = "default",
            *,
            expires_at: datetime | int | None = None,
            keep_ttl: bool = False) -> None:
        """Set an object into the collection"""
        if expires_at:
            expi = None

            if isinstance(expires_at, datetime):
                expires_at = int(expires_at.timestamp())

        elif keep_ttl:
            expi = None

        elif expires_in == "default":
            expi = self.default_expiration
        
        else:
            expi = expires_in

        expi = self.default_expiration if expires_in == "default" else expires_in
        expi = cast(int | None, expi)

        valdata = self._serialize_value(value)

        self.cdb.rclient.set(
            f"{self.key}:{key}", valdata,
            ex=expi, exat=expires_at, keepttl=keep_ttl
        )


    def get(self, key: str) -> CacheControllerModelT | None:
        """Get an object by its key from the collection"""
        rawval = self.cdb.rclient.get(f"{self.key}:{key}")
        return self._deserialize(rawval) # type: ignore


    def delete(self, key: str) -> None:
        """Delete an object by its key from the collection"""
        self.cdb.rclient.delete(f"{self.key}:{key}")


    remove = delete


    def pop(self, key: str) -> CacheControllerModelT | None:
        """Get an object then deletes it"""
        rawval = self.cdb.rclient.getdel(f"{self.key}:{key}")
        return self._deserialize(rawval) # type: ignore
    

    def getex(self,
              key: str,
              expires_in: int | None = None,
              *,
              expires_at: datetime | int | None = None,
              persist: bool = False):
        """Get an object then set an expiration"""
        rawval = self.cdb.rclient.getex(
            f"{self.key}:{key}",
            ex=expires_in,
            exat=expires_at,
            persist=persist
        )
        return self._deserialize(rawval) # type: ignore


    def getset(self,
               key: str,
               value: CacheControllerModelT) -> CacheControllerModelT | None:
        """Get an object then set its next value"""
        valdata = self._serialize_value(value)
        rawval = self.cdb.rclient.getset(f"{self.key}:{key}", valdata)
        return self._deserialize(rawval) # type: ignore


    def ttl(self, key: str) -> int | None:
        """Returns the time-to-live of an object"""
        return self.cdb.rclient.ttl(f"{self.key}:{key}") # type: ignore



from typing import TypeVar, Generic, Any, Self
from secrets import token_urlsafe
from datetime import datetime, timedelta

from pydantic import BaseModel, ConfigDict
import jwt
from jwt import PyJWK
from jwt.types import Options as JwtOptions
from jwt.algorithms import AllowedPublicKeys, AllowedPrivateKeys, get_default_algorithms

from .exceptions import JwtDenied


__all__ = ["JwtAuthConfig", "JwtAuthenticator"]


AuthDataT = TypeVar("AuthDataT", bound=BaseModel)



class JwtAuthConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    encodekey: AllowedPrivateKeys | PyJWK | str | bytes | None = None
    """The key used to encode (private) a JWTs (prefer using .key if symetric)"""
    decodekey: AllowedPublicKeys | PyJWK | str | bytes | None = None
    """The key used to decode (public) a JWTs (prefer using .key if symetric)"""
    key: PyJWK | str | bytes | None = None # symetric only
    """The key used to both encode and decode JWTs (symetric)"""
    algorithms: list[str] | None = None
    """The list of supported algorithms during JWT decoding"""
    issuer: str | None = None
    """JWT issuer (iss key)"""
    audience: list[str] | None = None
    """JWT audience (aud key)"""
    expiration: timedelta | int | None = None # TODO: or timedelta
    """JWT expiration in seconds or as a timedelta (exp key)"""



class JwtAuthenticator(Generic[AuthDataT]):
    """Base class to create an authentification method
    with JSON Web Tokens.
    """

    def __init__(self, data: AuthDataT) -> None:
        self.data = data
        """The object that was in the JWT"""


    def __init_subclass__(cls, model: type[AuthDataT]) -> None:
        cls._auth_data_model = model
        """The model of the payload in the JWT"""


    def __repr__(self) -> str:
        return "<{} data={!r}>".format(
            self.__class__.__name__,
            self.data
        )


    @staticmethod
    def verify_jwt(token: str) -> None:
        """A checker function that verifies the JWT
        before its decoding"""
        ...


    @staticmethod
    def verify_data(data: dict[str, Any]) -> None:
        """A chercker function that verifies the data
        before its validation by the model"""
        ...


    @classmethod
    def generate(cls,
                 obj: AuthDataT,
                 config: JwtAuthConfig,
                 extra_data: dict[str, Any] | None = None) -> str:
        """Generates a JSON Web Token with the given data
        and configuration"""
        iat = datetime.now()
        tokeninfo = {
            "jti": token_urlsafe(6),
            "iat": int(iat.timestamp())
        }

        if config.issuer is not None:
            tokeninfo["iss"] = config.issuer

        if config.audience is not None:
            tokeninfo["aud"] = config.audience

        if config.expiration is not None:
            exptd = config.expiration \
                if isinstance(config.expiration, timedelta) \
                else timedelta(seconds=config.expiration)
            tokeninfo["exp"] = int((iat + exptd).timestamp())

        encodekey = config.encodekey or config.key
        assert encodekey, "There is not either encodekey or key, you need to put one lil bro"

        payload = tokeninfo | obj.model_dump(mode="json") | (extra_data or {})
        jwtoken = jwt.encode(payload, encodekey)
        return jwtoken


    @classmethod
    def load(cls,
             token: str,
             config: JwtAuthConfig,
             opts: JwtOptions | None = None) -> Self:
        """Loads a JSON Web Token and returns an auth
        with the object inside the JWT payload as .data"""
        algorithms = config.algorithms
        if not algorithms:
            algorithms = list(get_default_algorithms().keys())

        decodekey = config.decodekey or config.key
        assert decodekey, "There is not either decodekey or key, you need to put one lil bro"

        try:
            cls.verify_jwt(token)
            data = jwt.decode(
                token,
                decodekey,
                algorithms=algorithms,
                options=opts,
                issuer=config.issuer
            )
        except jwt.PyJWTError as e:
            raise JwtDenied("Failed to verify JWT") from e

        try:
            cls.verify_data(data)
            authdata = cls._auth_data_model.model_validate(data, extra="ignore")
        except Exception as e:
            raise JwtDenied("The data from the given JWT is invalid") from e

        authenti = cls(authdata)
        return authenti
